# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: object_detection/protos/fpn.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='object_detection/protos/fpn.proto',
  package='object_detection.protos',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!object_detection/protos/fpn.proto\x12\x17object_detection.protos\"i\n\x16\x46\x65\x61turePyramidNetworks\x12\x14\n\tmin_level\x18\x01 \x01(\x05:\x01\x33\x12\x14\n\tmax_level\x18\x02 \x01(\x05:\x01\x37\x12#\n\x16\x61\x64\x64itional_layer_depth\x18\x03 \x01(\x05:\x03\x32\x35\x36\"\xa6\x01\n#BidirectionalFeaturePyramidNetworks\x12\x14\n\tmin_level\x18\x01 \x01(\x05:\x01\x33\x12\x14\n\tmax_level\x18\x02 \x01(\x05:\x01\x37\x12\x16\n\x0enum_iterations\x18\x03 \x01(\x05\x12\x13\n\x0bnum_filters\x18\x04 \x01(\x05\x12&\n\x0e\x63ombine_method\x18\x05 \x01(\t:\x0e\x66\x61st_attention'
)




_FEATUREPYRAMIDNETWORKS = _descriptor.Descriptor(
  name='FeaturePyramidNetworks',
  full_name='object_detection.protos.FeaturePyramidNetworks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_level', full_name='object_detection.protos.FeaturePyramidNetworks.min_level', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=3,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_level', full_name='object_detection.protos.FeaturePyramidNetworks.max_level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=7,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='additional_layer_depth', full_name='object_detection.protos.FeaturePyramidNetworks.additional_layer_depth', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=256,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=167,
)


_BIDIRECTIONALFEATUREPYRAMIDNETWORKS = _descriptor.Descriptor(
  name='BidirectionalFeaturePyramidNetworks',
  full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_level', full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks.min_level', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=3,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_level', full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks.max_level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=7,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_iterations', full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks.num_iterations', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_filters', full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks.num_filters', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='combine_method', full_name='object_detection.protos.BidirectionalFeaturePyramidNetworks.combine_method', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"fast_attention".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=170,
  serialized_end=336,
)

DESCRIPTOR.message_types_by_name['FeaturePyramidNetworks'] = _FEATUREPYRAMIDNETWORKS
DESCRIPTOR.message_types_by_name['BidirectionalFeaturePyramidNetworks'] = _BIDIRECTIONALFEATUREPYRAMIDNETWORKS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FeaturePyramidNetworks = _reflection.GeneratedProtocolMessageType('FeaturePyramidNetworks', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREPYRAMIDNETWORKS,
  '__module__' : 'object_detection.protos.fpn_pb2'
  # @@protoc_insertion_point(class_scope:object_detection.protos.FeaturePyramidNetworks)
  })
_sym_db.RegisterMessage(FeaturePyramidNetworks)

BidirectionalFeaturePyramidNetworks = _reflection.GeneratedProtocolMessageType('BidirectionalFeaturePyramidNetworks', (_message.Message,), {
  'DESCRIPTOR' : _BIDIRECTIONALFEATUREPYRAMIDNETWORKS,
  '__module__' : 'object_detection.protos.fpn_pb2'
  # @@protoc_insertion_point(class_scope:object_detection.protos.BidirectionalFeaturePyramidNetworks)
  })
_sym_db.RegisterMessage(BidirectionalFeaturePyramidNetworks)


# @@protoc_insertion_point(module_scope)
